@@ -22,6 +22,9 @@
  */
 
 #include "config.h"
+
+#include <string.h>
+
 #include <glib/gi18n-lib.h>
 #include <stdlib.h>
 
@@ -72,15 +75,38 @@ enum
 
 static gboolean goa_imap_auth_login_is_needed (GoaMailAuth        *auth);
 static gboolean goa_imap_auth_login_run_sync (GoaMailAuth         *_auth,
-                                              GDataInputStream    *input,
-                                              GDataOutputStream   *output,
                                               GCancellable        *cancellable,
                                               GError             **error);
+static gboolean goa_imap_auth_login_starttls_sync (GoaMailAuth    *_auth,
+                                                   GCancellable   *cancellable,
+                                                   GError        **error);
 
 G_DEFINE_TYPE (GoaImapAuthLogin, goa_imap_auth_login, GOA_TYPE_MAIL_AUTH);
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+static gchar *
+imap_auth_escape_backslash (const gchar *str)
+{
+  GString *ret;
+  gsize i;
+  gsize len;
+
+  ret = g_string_new ("");
+  len = strlen (str);
+
+  for (i = 0; i < len; i++)
+    {
+      if (str[i] == '\\')
+        g_string_append_c (ret, '\\');
+      g_string_append_c (ret, str[i]);
+    }
+
+  return g_string_free (ret, FALSE);
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
 static void
 goa_imap_auth_login_finalize (GObject *object)
 {
@@ -180,6 +206,7 @@ goa_imap_auth_login_class_init (GoaImapAuthLoginClass *klass)
   auth_class = GOA_MAIL_AUTH_CLASS (klass);
   auth_class->is_needed = goa_imap_auth_login_is_needed;
   auth_class->run_sync = goa_imap_auth_login_run_sync;
+  auth_class->starttls_sync = goa_imap_auth_login_starttls_sync;
 
   /**
    * GoaImapAuthLogin:provider:
@@ -299,12 +326,12 @@ goa_imap_auth_login_is_needed (GoaMailAuth *_auth)
 
 static gboolean
 goa_imap_auth_login_run_sync (GoaMailAuth         *_auth,
-                              GDataInputStream    *input,
-                              GDataOutputStream   *output,
                               GCancellable        *cancellable,
                               GError             **error)
 {
   GoaImapAuthLogin *auth = GOA_IMAP_AUTH_LOGIN (_auth);
+  GDataInputStream *input;
+  GDataOutputStream *output;
   gchar *request;
   gchar *response;
   gboolean ret;
@@ -317,11 +344,13 @@ goa_imap_auth_login_run_sync (GoaMailAuth         *_auth,
 
   if (auth->password != NULL)
     {
-      password = g_strdup (auth->password);
+      password = imap_auth_escape_backslash (auth->password);
     }
   else if (auth->provider != NULL && auth->object != NULL)
     {
       GVariant *credentials;
+      gchar *value;
+
       credentials = goa_utils_lookup_credentials_sync (auth->provider,
                                                        auth->object,
                                                        cancellable,
@@ -331,7 +360,7 @@ goa_imap_auth_login_run_sync (GoaMailAuth         *_auth,
           g_prefix_error (error, "Error looking up credentials for IMAP LOGIN in keyring: ");
           goto out;
         }
-      if (!g_variant_lookup (credentials, "imap-password", "s", &password))
+      if (!g_variant_lookup (credentials, "imap-password", "s", &value))
         {
           g_set_error (error,
                        GOA_ERROR,
@@ -340,6 +369,10 @@ goa_imap_auth_login_run_sync (GoaMailAuth         *_auth,
           g_variant_unref (credentials);
           goto out;
         }
+
+      password = imap_auth_escape_backslash (value);
+
+      g_free (value);
       g_variant_unref (credentials);
     }
   else
@@ -351,7 +384,11 @@ goa_imap_auth_login_run_sync (GoaMailAuth         *_auth,
       goto out;
     }
 
+  input = goa_mail_auth_get_input (_auth);
+  output = goa_mail_auth_get_output (_auth);
+
   request = g_strdup_printf ("A001 LOGIN \"%s\" \"%s\"\r\n", auth->username, password);
+  g_debug ("> A001 LOGIN \"********************\" \"********************\"");
   if (!g_data_output_stream_put_string (output, request, cancellable, error))
     goto out;
 
@@ -359,6 +396,62 @@ goa_imap_auth_login_run_sync (GoaMailAuth         *_auth,
   response = g_data_input_stream_read_line (input, NULL, cancellable, error);
   if (response == NULL)
     goto out;
+  g_debug ("< %s", response);
+  /* ignore untagged responses */
+  if (g_str_has_prefix (response, "*"))
+    {
+      g_free (response);
+      goto again;
+    }
+  if (!g_str_has_prefix (response, "A001 OK"))
+    {
+      g_set_error (error,
+                   GOA_ERROR,
+                   GOA_ERROR_FAILED,
+                   _("Authentication failed"));
+      goto out;
+    }
+
+  ret = TRUE;
+
+ out:
+  g_free (response);
+  g_free (request);
+  g_free (password);
+  return ret;
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+static gboolean
+goa_imap_auth_login_starttls_sync (GoaMailAuth         *_auth,
+                                   GCancellable        *cancellable,
+                                   GError             **error)
+{
+  GDataInputStream *input;
+  GDataOutputStream *output;
+  gchar *request;
+  gchar *response;
+  gboolean ret;
+
+  request = NULL;
+  response = NULL;
+
+  ret = FALSE;
+
+  input = goa_mail_auth_get_input (_auth);
+  output = goa_mail_auth_get_output (_auth);
+
+  request = g_strdup ("A001 STARTTLS\r\n");
+  g_debug ("> %s", request);
+  if (!g_data_output_stream_put_string (output, request, cancellable, error))
+    goto out;
+
+ again:
+  response = g_data_input_stream_read_line (input, NULL, cancellable, error);
+  if (response == NULL)
+    goto out;
+  g_debug ("< %s", response);
   /* ignore untagged responses */
   if (g_str_has_prefix (response, "*"))
     {
@@ -380,6 +473,5 @@ goa_imap_auth_login_run_sync (GoaMailAuth         *_auth,
  out:
   g_free (response);
   g_free (request);
-  g_free (password);
   return ret;
 }
