@@ -54,6 +54,7 @@ struct _GoaSmtpAuthPlain
   GoaProvider *provider;
   GoaObject *object;
   gboolean auth_supported;
+  gboolean greeting_absent;
   gchar *domain;
   gchar *username;
   gchar *password;
@@ -77,15 +78,199 @@ enum
 
 static gboolean goa_smtp_auth_plain_is_needed (GoaMailAuth        *_auth);
 static gboolean goa_smtp_auth_plain_run_sync (GoaMailAuth         *_auth,
-                                              GDataInputStream    *input,
-                                              GDataOutputStream   *output,
                                               GCancellable        *cancellable,
                                               GError             **error);
+static gboolean goa_smtp_auth_plain_starttls_sync (GoaMailAuth    *_auth,
+                                                   GCancellable   *cancellable,
+                                                   GError        **error);
 
 G_DEFINE_TYPE (GoaSmtpAuthPlain, goa_smtp_auth_plain, GOA_TYPE_MAIL_AUTH);
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+static gboolean
+smtp_auth_plain_check_not_220 (const gchar *response, GError **error)
+{
+  if (!g_str_has_prefix (response, "220"))
+    {
+      g_set_error (error,
+                   GOA_ERROR,
+                   GOA_ERROR_FAILED, /* TODO: more specific */
+                   "Unexpected response `%s' while doing PLAIN authentication",
+                   response);
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
+static gboolean
+smtp_auth_plain_check_not_235 (const gchar *response, GError **error)
+{
+  if (!g_str_has_prefix (response, "235"))
+    {
+      g_set_error (error,
+                   GOA_ERROR,
+                   GOA_ERROR_FAILED, /* TODO: more specific */
+                   _("Authentication failed"));
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
+static gboolean
+smtp_auth_plain_check_not_250 (const gchar *response, GError **error)
+{
+  if (!g_str_has_prefix (response, "250") || strlen (response) < 4)
+    {
+      g_set_error (error,
+                   GOA_ERROR,
+                   GOA_ERROR_FAILED, /* TODO: more specific */
+                   "Unexpected response `%s' while doing PLAIN authentication",
+                   response);
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
+static gboolean
+smtp_auth_plain_check_421 (const gchar *response, GError **error)
+{
+  if (g_str_has_prefix (response, "421"))
+    {
+      g_set_error (error,
+                   GOA_ERROR,
+                   GOA_ERROR_FAILED, /* TODO: more specific */
+                   _("Service not available"));
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
+static gboolean
+smtp_auth_plain_check_454 (const gchar *response, GError **error)
+{
+  if (g_str_has_prefix (response, "454"))
+    {
+      g_set_error (error,
+                   GOA_ERROR,
+                   GOA_ERROR_FAILED, /* TODO: more specific */
+                   _("TLS not available"));
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+static gchar *
+smtp_auth_plain_get_domain (GoaSmtpAuthPlain  *auth,
+                            GError           **error)
+{
+  gchar *domain;
+
+  domain = NULL;
+
+  if (auth->domain != NULL)
+    {
+      domain = g_strdup (auth->domain);
+    }
+  else if (auth->object != NULL)
+    {
+      GoaMail *mail;
+      gchar *email_address;
+
+      mail = goa_object_get_mail (auth->object);
+      if (mail == NULL)
+        {
+          g_set_error (error,
+                       GOA_ERROR,
+                       GOA_ERROR_FAILED, /* TODO: more specific */
+                       _("org.gnome.OnlineAccounts.Mail is not available"));
+          goto out;
+        }
+
+      email_address = goa_mail_dup_email_address (mail);
+      if (!goa_utils_parse_email_address (email_address, NULL, &domain))
+        {
+          g_set_error (error,
+                       GOA_ERROR,
+                       GOA_ERROR_FAILED, /* TODO: more specific */
+                       _("Failed to parse email address"));
+          goto out;
+        }
+
+      g_free (email_address);
+      g_object_unref (mail);
+    }
+  else
+    {
+      g_set_error (error,
+                   GOA_ERROR,
+                   GOA_ERROR_FAILED, /* TODO: more specific */
+                   _("Cannot do SMTP PLAIN without a domain"));
+      goto out;
+    }
+
+ out:
+  return domain;
+}
+
+static gchar *
+smtp_auth_plain_get_password (GoaSmtpAuthPlain  *auth,
+                              GCancellable      *cancellable,
+                              GError           **error)
+{
+  gchar *password;
+
+  password = NULL;
+
+  if (auth->password != NULL)
+    {
+      password = g_strdup (auth->password);
+    }
+  else if (auth->provider != NULL && auth->object != NULL)
+    {
+      GVariant *credentials;
+      credentials = goa_utils_lookup_credentials_sync (auth->provider,
+                                                       auth->object,
+                                                       cancellable,
+                                                       error);
+      if (credentials == NULL)
+        {
+          g_prefix_error (error, "Error looking up credentials for SMTP PLAIN in keyring: ");
+          goto out;
+        }
+      if (!g_variant_lookup (credentials, "smtp-password", "s", &password))
+        {
+          g_set_error (error,
+                       GOA_ERROR,
+                       GOA_ERROR_FAILED, /* TODO: more specific */
+                       _("Did not find smtp-password in credentials"));
+          g_variant_unref (credentials);
+          goto out;
+        }
+      g_variant_unref (credentials);
+    }
+  else
+    {
+      g_set_error (error,
+                   GOA_ERROR,
+                   GOA_ERROR_FAILED, /* TODO: more specific */
+                   _("Cannot do SMTP PLAIN without a password"));
+      goto out;
+    }
+
+ out:
+  return password;
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
 static void
 goa_smtp_auth_plain_finalize (GObject *object)
 {
@@ -194,6 +379,7 @@ goa_smtp_auth_plain_class_init (GoaSmtpAuthPlainClass *klass)
   auth_class = GOA_MAIL_AUTH_CLASS (klass);
   auth_class->is_needed = goa_smtp_auth_plain_is_needed;
   auth_class->run_sync = goa_smtp_auth_plain_run_sync;
+  auth_class->starttls_sync = goa_smtp_auth_plain_starttls_sync;
 
   /**
    * GoaSmtpAuthPlain:provider:
@@ -336,12 +522,12 @@ goa_smtp_auth_plain_is_needed (GoaMailAuth *_auth)
 
 static gboolean
 goa_smtp_auth_plain_run_sync (GoaMailAuth         *_auth,
-                              GDataInputStream    *input,
-                              GDataOutputStream   *output,
                               GCancellable        *cancellable,
                               GError             **error)
 {
   GoaSmtpAuthPlain *auth = GOA_SMTP_AUTH_PLAIN (_auth);
+  GDataInputStream *input;
+  GDataOutputStream *output;
   gboolean plain_supported;
   gboolean ret;
   gchar *auth_arg_base64;
@@ -362,107 +548,31 @@ goa_smtp_auth_plain_run_sync (GoaMailAuth         *_auth,
 
   ret = FALSE;
 
-  if (auth->password != NULL)
-    {
-      password = g_strdup (auth->password);
-    }
-  else if (auth->provider != NULL && auth->object != NULL)
-    {
-      GVariant *credentials;
-      credentials = goa_utils_lookup_credentials_sync (auth->provider,
-                                                       auth->object,
-                                                       cancellable,
-                                                       error);
-      if (credentials == NULL)
-        {
-          g_prefix_error (error, "Error looking up credentials for SMTP PLAIN in keyring: ");
-          goto out;
-        }
-      if (!g_variant_lookup (credentials, "smtp-password", "s", &password))
-        {
-          g_set_error (error,
-                       GOA_ERROR,
-                       GOA_ERROR_FAILED, /* TODO: more specific */
-                       _("Did not find smtp-password in credentials"));
-          g_variant_unref (credentials);
-          goto out;
-        }
-      g_variant_unref (credentials);
-    }
-  else
-    {
-      g_set_error (error,
-                   GOA_ERROR,
-                   GOA_ERROR_FAILED, /* TODO: more specific */
-                   _("Cannot do SMTP PLAIN without a password"));
-      goto out;
-    }
-
-  if (auth->domain != NULL)
-    {
-      domain = g_strdup (auth->domain);
-    }
-  else if (auth->object != NULL)
-    {
-      GoaMail *mail;
-      gchar *email_address;
-
-      mail = goa_object_get_mail (auth->object);
-      if (mail == NULL)
-        {
-          g_set_error (error,
-                       GOA_ERROR,
-                       GOA_ERROR_FAILED, /* TODO: more specific */
-                       _("org.gnome.OnlineAccounts.Mail is not available"));
-          goto out;
-        }
-
-      email_address = goa_mail_dup_email_address (mail);
-      if (!goa_utils_parse_email_address (email_address, NULL, &domain))
-        {
-          g_set_error (error,
-                       GOA_ERROR,
-                       GOA_ERROR_FAILED, /* TODO: more specific */
-                       _("Failed to parse email address"));
-          goto out;
-        }
-
-      g_free (email_address);
-      g_object_unref (mail);
-    }
-  else
-    {
-      g_set_error (error,
-                   GOA_ERROR,
-                   GOA_ERROR_FAILED, /* TODO: more specific */
-                   _("Cannot do SMTP PLAIN without a domain"));
-      goto out;
-    }
-
-  /* Check the greeting */
-
-  response = g_data_input_stream_read_line (input, NULL, cancellable, error);
-  if (response == NULL)
+  password = smtp_auth_plain_get_password (auth, cancellable, error);
+  if (password == NULL)
     goto out;
-  g_debug ("< %s", response);
-  if (g_str_has_prefix (response, "421"))
+
+  domain = smtp_auth_plain_get_domain (auth, error);
+  if (domain == NULL)
+    goto out;
+
+  input = goa_mail_auth_get_input (_auth);
+  output = goa_mail_auth_get_output (_auth);
+
+  /* Check the greeting, if there is one */
+
+  if (!auth->greeting_absent)
     {
-      g_set_error (error,
-                   GOA_ERROR,
-                   GOA_ERROR_FAILED, /* TODO: more specific */
-                   _("Service not available"));
-      goto out;
+      response = g_data_input_stream_read_line (input, NULL, cancellable, error);
+      if (response == NULL)
+        goto out;
+      g_debug ("< %s", response);
+      if (smtp_auth_plain_check_421 (response, error))
+        goto out;
+      if (smtp_auth_plain_check_not_220 (response, error))
+        goto out;
+      g_clear_pointer (&response, g_free);
     }
-  if (!g_str_has_prefix (response, "220"))
-    {
-      g_set_error (error,
-                   GOA_ERROR,
-                   GOA_ERROR_FAILED, /* TODO: more specific */
-                   "Unexpected response `%s' while doing PLAIN authentication",
-                   response);
-      goto out;
-    }
-  g_clear_pointer (&response, g_free);
 
   /* Send EHLO */
 
@@ -479,23 +589,10 @@ goa_smtp_auth_plain_run_sync (GoaMailAuth         *_auth,
   if (response == NULL)
     goto out;
   g_debug ("< %s", response);
-  if (g_str_has_prefix (response, "421"))
-    {
-      g_set_error (error,
-                   GOA_ERROR,
-                   GOA_ERROR_FAILED, /* TODO: more specific */
-                   _("Service not available"));
-      goto out;
-    }
-  if (!g_str_has_prefix (response, "250") || strlen (response) < 4)
-    {
-      g_set_error (error,
-                   GOA_ERROR,
-                   GOA_ERROR_FAILED, /* TODO: more specific */
-                   "Unexpected response `%s' while doing PLAIN authentication",
-                   response);
-      goto out;
-    }
+  if (smtp_auth_plain_check_421 (response, error))
+    goto out;
+  if (smtp_auth_plain_check_not_250 (response, error))
+    goto out;
 
   if (g_str_has_prefix (response + 4, "AUTH"))
     {
@@ -531,7 +628,7 @@ goa_smtp_auth_plain_run_sync (GoaMailAuth         *_auth,
   auth_arg_base64 = g_base64_encode ((guchar *) auth_arg_plain, auth_arg_plain_len);
 
   request = g_strdup_printf ("AUTH PLAIN %s\r\n", auth_arg_base64);
-  g_debug ("> AUTH PLAN ********************");
+  g_debug ("> AUTH PLAIN ********************");
   if (!g_data_output_stream_put_string (output, request, cancellable, error))
     goto out;
   g_clear_pointer (&request, g_free);
@@ -540,14 +637,8 @@ goa_smtp_auth_plain_run_sync (GoaMailAuth         *_auth,
   if (response == NULL)
     goto out;
   g_debug ("< %s", response);
-  if (!g_str_has_prefix (response, "235"))
-    {
-      g_set_error (error,
-                   GOA_ERROR,
-                   GOA_ERROR_FAILED, /* TODO: more specific */
-                   _("Authentication failed"));
-      goto out;
-    }
+  if (smtp_auth_plain_check_not_235 (response, error))
+    goto out;
   g_clear_pointer (&response, g_free);
 
   ret = TRUE;
@@ -561,3 +652,113 @@ goa_smtp_auth_plain_run_sync (GoaMailAuth         *_auth,
   g_free (request);
   return ret;
 }
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+static gboolean
+goa_smtp_auth_plain_starttls_sync (GoaMailAuth         *_auth,
+                                   GCancellable        *cancellable,
+                                   GError             **error)
+{
+  GoaSmtpAuthPlain *auth = GOA_SMTP_AUTH_PLAIN (_auth);
+  GDataInputStream *input;
+  GDataOutputStream *output;
+  gboolean ret;
+  gboolean starttls_supported;
+  gchar *domain;
+  gchar *request;
+  gchar *response;
+
+  starttls_supported = FALSE;
+  domain = NULL;
+  request = NULL;
+  response = NULL;
+
+  ret = FALSE;
+
+  domain = smtp_auth_plain_get_domain (auth, error);
+  if (domain == NULL)
+    goto out;
+
+  input = goa_mail_auth_get_input (_auth);
+  output = goa_mail_auth_get_output (_auth);
+
+  /* Check the greeting */
+
+  response = g_data_input_stream_read_line (input, NULL, cancellable, error);
+  if (response == NULL)
+    goto out;
+  g_debug ("< %s", response);
+  if (smtp_auth_plain_check_421 (response, error))
+    goto out;
+  if (smtp_auth_plain_check_not_220 (response, error))
+    goto out;
+  g_clear_pointer (&response, g_free);
+
+  /* Send EHLO */
+
+  request = g_strdup_printf ("EHLO %s\r\n", domain);
+  g_debug ("> %s", request);
+  if (!g_data_output_stream_put_string (output, request, cancellable, error))
+    goto out;
+  g_clear_pointer (&request, g_free);
+
+  /* Check if STARTTLS is supported or not */
+
+ ehlo_again:
+  response = g_data_input_stream_read_line (input, NULL, cancellable, error);
+  if (response == NULL)
+    goto out;
+  g_debug ("< %s", response);
+  if (smtp_auth_plain_check_421 (response, error))
+    goto out;
+  if (smtp_auth_plain_check_not_250 (response, error))
+    goto out;
+
+  if (g_str_has_prefix (response + 4, "STARTTLS"))
+    starttls_supported = TRUE;
+
+  if (response[3] == '-')
+    {
+      g_free (response);
+      goto ehlo_again;
+    }
+  else if (!starttls_supported)
+    {
+      g_set_error (error,
+                   GOA_ERROR,
+                   GOA_ERROR_NOT_SUPPORTED,
+                   _("Server does not support STARTTLS"));
+      goto out;
+    }
+  g_clear_pointer (&response, g_free);
+
+  /* Send STARTTLS */
+
+  request = g_strdup ("STARTTLS\r\n");
+  g_debug ("> %s", request);
+  if (!g_data_output_stream_put_string (output, request, cancellable, error))
+    goto out;
+  g_clear_pointer (&request, g_free);
+
+  response = g_data_input_stream_read_line (input, NULL, cancellable, error);
+  if (response == NULL)
+    goto out;
+  g_debug ("< %s", response);
+  if (smtp_auth_plain_check_454 (response, error))
+    goto out;
+  if (smtp_auth_plain_check_not_220 (response, error))
+    goto out;
+  g_clear_pointer (&response, g_free);
+
+  /* There won't be a greeting after this */
+  auth->greeting_absent = TRUE;
+
+  ret = TRUE;
+
+ out:
+  g_free (domain);
+  g_free (response);
+  g_free (request);
+  return ret;
+}
